package java.util.concurrent.atomic;

import sun.misc.Unsafe;

public class AtomicMarkableReference<V>
{
    private volatile AtomicMarkableReference.Pair<V> pair;
    private static final Unsafe UNSAFE;
    private static final long pairOffset;
    
    public AtomicMarkableReference(V v, boolean b) {
        this.pair = AtomicMarkableReference.Pair.of(v, b);
    }
    
    public V getReference() {
        return this.pair.reference;
    }
    
    public boolean isMarked() {
        return this.pair.mark;
    }
    
    public V get(boolean[] array) {
        AtomicMarkableReference.Pair<Object> pair = this.pair;
        array[0] = pair.mark;
        return pair.reference;
    }
    
    public boolean weakCompareAndSet(V v, V v2, boolean b, boolean b2) {
        return this.compareAndSet(v, v2, b, b2);
    }
    
    public boolean compareAndSet(V v, V v2, boolean b, boolean b2) {
        AtomicMarkableReference.Pair<V> pair = this.pair;
        return v == pair.reference && b == pair.mark && ((v2 == pair.reference && b2 == pair.mark) || this.casPair(pair, AtomicMarkableReference.Pair.of(v2, b2)));
    }
    
    public void set(V v, boolean b) {
        AtomicMarkableReference.Pair<Object> pair = this.pair;
        if (v != pair.reference || b != pair.mark) {
            this.pair = AtomicMarkableReference.Pair.of(v, b);
        }
    }
    
    public boolean attemptMark(V v, boolean b) {
        AtomicMarkableReference.Pair<V> pair = this.pair;
        return v == pair.reference && (b == pair.mark || this.casPair(pair, AtomicMarkableReference.Pair.of(v, b)));
    }
    
    private boolean casPair(AtomicMarkableReference.Pair<V> expected, AtomicMarkableReference.Pair<V> x) {
        return AtomicMarkableReference.UNSAFE.compareAndSwapObject(this, AtomicMarkableReference.pairOffset, expected, x);
    }
    
    static long objectFieldOffset(Unsafe unsafe, String s, Class<?> clazz) {
        try {
            return unsafe.objectFieldOffset(clazz.getDeclaredField(s));
        }
        catch (NoSuchFieldException cause) {
            NoSuchFieldError noSuchFieldError = new NoSuchFieldError(s);
            noSuchFieldError.initCause(cause);
            throw noSuchFieldError;
        }
    }
    
    static {
        UNSAFE = Unsafe.getUnsafe();
        pairOffset = objectFieldOffset(AtomicMarkableReference.UNSAFE, "pair", AtomicMarkableReference.class);
    }
    
    private static class Pair<T>
    {
        final T reference;
        final boolean mark;
        
        private Pair(T reference, boolean mark) {
            this.reference = reference;
            this.mark = mark;
        }
        
        static <T> AtomicMarkableReference.Pair<T> of(T t, boolean b) {
            return new AtomicMarkableReference.Pair(t, b);
        }
    }
}
